/* 
 * $Id: nexthop.c,v 1.2 1998/11/23 22:49:16 masaki Exp $
 */

#include <mrt.h>


static u_int
nexthop_hash_fn (prefix_t *prefix, u_int size)
{
#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6) {
	u_int val, buff[4];
	memcpy (buff, prefix_tochar (prefix), 16);
	val = buff[0] ^ buff[1] ^ buff[2] ^ buff[3];
	val ^= (val >> 16);
	val = val % size;
	return (val);
    }
    else
#endif /* HAVE_IPV6 */
    if (prefix->family == AF_INET) {
	u_int val;
	u_char dest[4];
	memcpy (dest, prefix_tochar (prefix), 4);
	val = dest[0] + dest[1] + dest[2] + dest[3];
	val = val % size;
	return (val);
    }
    else {
	assert (0);
    }
    /* NEVER REACHES */
    return (0);
}


static u_int
nexthop_lookup_fn (prefix_t *a, prefix_t *b)
{
    return (prefix_compare2 (a, b) == 0);
}


nexthop_t *
ref_nexthop (nexthop_t *nexthop)
{
    if (nexthop) {
        pthread_mutex_lock (&nexthop->mutex_lock);
        assert (nexthop->ref_count > 0);
        nexthop->ref_count++;
        pthread_mutex_unlock (&nexthop->mutex_lock);
    }
    return (nexthop);
}


/* These tables are to store interface nexthops (address is zeros) */

static nexthop_t *interface_nexthop[MAX_INTERFACES];
#ifdef HAVE_IPV6
static nexthop_t *interface_nexthop6[MAX_INTERFACES];
#endif /* HAVE_IPV6 */


#define NEXTHOP_HASH_SIZE 1023
/* 
 * find nexthop or create a new one if it does not exit
 */
nexthop_t *
add_nexthop (prefix_t *prefix, interface_t *interface)
{
    nexthop_t nh, *nexthop;
    mrt_hash_table_t *hash = &MRT->hash_table;

    if (prefix == NULL)
	return (NULL);

#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
	hash = &MRT->hash_table6;
#endif /* HAVE_IPV6 */

    pthread_mutex_lock (&hash->mutex_lock);

    if (prefix_is_unspecified (prefix)) {
	nexthop_t **nexthop_table = interface_nexthop;
	int index = (interface)? interface->index: 0;

#ifdef HAVE_IPV6
        if (prefix->family == AF_INET6)
	    nexthop_table = interface_nexthop6;
#endif /* HAVE_IPV6 */
	if ((nexthop = nexthop_table[index]) == NULL) {
	    nexthop = nexthop_table[index] = New (nexthop_t);
	    nexthop->prefix = Ref_Prefix (prefix);
	    nexthop->interface = interface;
	    nexthop->ref_count = 1;
    	    pthread_mutex_init (&nexthop->mutex_lock, NULL);
            pthread_mutex_unlock (&hash->mutex_lock);
	    return (nexthop);
	}
	nexthop = ref_nexthop (nexthop);
        pthread_mutex_unlock (&hash->mutex_lock);
	return (nexthop);
    }

    if (hash->table == NULL) {
        hash->table = HASH_Create (NEXTHOP_HASH_SIZE, 
		    HASH_KeyOffset, HASH_Offset (&nh, &nh.prefix),
                    HASH_LookupFunction, nexthop_lookup_fn,
                    HASH_HashFunction, nexthop_hash_fn, NULL);
    }

    if ((nexthop = HASH_Lookup (hash->table, prefix))) {
	nexthop = ref_nexthop (nexthop);
	if (nexthop->interface != interface && interface != NULL) {
	    /* can not reset the interface with NULL */
            trace ((nexthop->interface)? TR_WARN: TR_STATE, MRT->trace, 
		   "add_nexthop: %s on %s -> %s (interface change)\n",
                   prefix_toa (nexthop->prefix),
                   (nexthop->interface)?nexthop->interface->name:"?",
                   interface->name);
	    if (nexthop->interface == NULL)
	        nexthop->interface = interface;
	}
        pthread_mutex_unlock (&hash->mutex_lock);
	return (nexthop);
    }

    nexthop = New (nexthop_t);
    nexthop->prefix = Ref_Prefix (prefix);
    nexthop->interface = interface;
    nexthop->ref_count = 1;
    pthread_mutex_init (&nexthop->mutex_lock, NULL);
    HASH_Insert (hash->table, nexthop);
    pthread_mutex_unlock (&hash->mutex_lock);
    trace (TR_STATE, MRT->trace, "add_nexthop: %s on %s (new)\n",
           prefix_toa (nexthop->prefix), (interface)?interface->name:"?");
    return (nexthop);
}


#ifdef notdef
nexthop_t *
find_nexthop (prefix_t *prefix, interface_t *interface)
{
    nexthop_t nh, *nexthop = &nh;
    mrt_hash_table_t *hash = &MRT->hash_table;

    if (prefix == NULL)
	return (NULL);

#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
	hash = &MRT->hash_table6;
#endif /* HAVE_IPV6 */

    if (hash->table == NULL)
	return (NULL);

    pthread_mutex_lock (&hash->mutex_lock);

    nexthop->prefix = prefix;
    nexthop->interface = interface;

    if ((nexthop = HASH_Lookup (hash->table, nexthop))) {
        pthread_mutex_unlock (&hash->mutex_lock);
	return (nexthop);
    }
    return (NULL);
}
#endif


void
deref_nexthop (nexthop_t *nexthop)
{
    if (nexthop == NULL)
	return;
    pthread_mutex_lock (&nexthop->mutex_lock);
    assert (nexthop->ref_count > 0);
    if (nexthop->ref_count <= 1) {
	mrt_hash_table_t *hash = &MRT->hash_table;
#ifdef HAVE_IPV6
	if (nexthop->prefix->family == AF_INET6)
	    hash = &MRT->hash_table6;
#endif /* HAVE_IPV6 */
        pthread_mutex_lock (&hash->mutex_lock);
        /* someone may be searching in the table 
	       and found this at the same time */
	if (nexthop->ref_count <= 1) {
	    HASH_Remove (hash->table, nexthop);
            pthread_mutex_unlock (&hash->mutex_lock);
	    pthread_mutex_destroy (&nexthop->mutex_lock);
	    Deref_Prefix (nexthop->prefix);
	    Delete (nexthop);
	    return;
	}
        pthread_mutex_unlock (&hash->mutex_lock);
    }
    nexthop->ref_count--;
    pthread_mutex_unlock (&nexthop->mutex_lock);
    return;
}

